import os
import cv2

os.environ['OPENCV_FFMPEG_DEBUG'] = '0'
os.environ['OPENCV_VIDEOIO_DEBUG'] = '0'


class Video:
    @staticmethod
    def fill_images(video_file, frames):
        cam = cv2.VideoCapture(video_file)
        frame_count = cam.get(cv2.CAP_PROP_FRAME_COUNT)
        for i in range(int(frame_count)):
            ret, frame = cam.read()
            if ret:
                frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))

    @staticmethod
    def get_video_duration(video_path):
        video = cv2.VideoCapture(video_path)
        frame = video.get(cv2.CAP_PROP_FRAME_COUNT)
        fps = video.get(cv2.CAP_PROP_FPS)
        duration_in_seconds = frame / fps
        return duration_in_seconds



import numpy as np
import cv2


class Line:
    def __init__(self, coords):
        self.__pts = np.array(coords, np.int32).reshape((- 1, 1, 2))
        self.__rect = cv2.boundingRect(self.__pts)
        self.__mask = None
        self.__count_mask_pixels = 0
        self.__color = (255, 255, 255)
        self.__final_size = (256, 256)
        self.__final_orig_size = (256, 256, 3)

    def __get_mask(self, shape):
        if self.__mask is not None:
            return self.__mask
        self.__mask = np.full((shape[0], shape[1]), 0, dtype=np.uint8)
        cv2.fillPoly(self.__mask, [self.__pts], self.__color)
        x, y, w, h = self.__rect
        self.__mask = self.__mask[y:y + h, x:x + w]
        self.__count_mask_pixels = cv2.countNonZero(self.__mask)
        return self.__mask

    def get_dataset_images(self, orig_image, res_image):
        return self.get_image(orig_image), self.get_image(res_image, False)

    def get_image(self, image, orig=True):
        mask = self.__get_mask(image.shape)
        x, y, w, h = self.__rect
        orig_crop = image[y:y + h, x:x + w]
        final_image = np.full(self.__final_orig_size if orig is True else self.__final_size, 0, dtype=np.uint8)
        final_image[0:h, 0:w] = cv2.bitwise_or(orig_crop, orig_crop, mask=mask)
        return final_image

    def get_tlcr(self, image):
        return cv2.sumElems(image)[0] / self.__count_mask_pixels

    def sum_elems_optimized(self, mask):
        pixels = np.argwhere(mask == 255)
        x_dict = {}
        x_values = []
        for coords in pixels:
            if str(coords[1]) not in x_dict:
                x_dict[str(coords[1])] = []
                x_values.append(coords[1])
            x_dict[str(coords[1])].append(coords[0])
        x_values.sort()
        x_bound, y_bound, _, _ = self.__rect
        new_pts = []
        for i in range(4):
            new_pts.append([self.__pts[i][0][0] - x_bound, self.__pts[i][0][1] - y_bound])

        #sort new_points
        new_pts.sort(key=lambda x: x[1])

        if new_pts[0][0] > new_pts[1][0]:
            self.swap_positions(new_pts, 0, 1)

        if new_pts[2][0] > new_pts[3][0]:
            self.swap_positions(new_pts, 2, 3)

        height_min = [0, 0]
        pixel_stripes = []
        pixel_stripes.append([])
        for i in range(new_pts[2][0], new_pts[3][0] + 1):
            popped = x_dict[str(i)].pop()
            pixel_stripes[0].append([i, popped])
            if i == new_pts[2][0]:
                height_min[0] = popped - 1
            elif i == new_pts[3][0]:
                height_min[1] = popped - 1

        reached_checkpoint = False
        met_height_value = []
        met_height_value.append(False)
        met_height_value.append(False)
        stripe_ind = 0
        while not reached_checkpoint:
            met_height_value[0] = False
            met_height_value[1] = False
            stripe_ind += 1
            for x in x_values:
                if len(x_dict[str(x)]) > 0:
                    if (x == new_pts[0][0] and x_dict[str(x)][len(x_dict[str(x)]) - 1] == new_pts[0][1])\
                            or (x == new_pts[1][0] and x_dict[str(x)][len(x_dict[str(x)]) - 1] == new_pts[1][1]):
                        reached_checkpoint = True

                    if height_min[0] == x_dict[str(x)][len(x_dict[str(x)]) - 1] and not met_height_value[0]:
                        pixel_stripes.append([])
                        met_height_value[0] = True
                        height_min[0] -= 1

                    if height_min[1] == x_dict[str(x)][len(x_dict[str(x)]) - 1] and not met_height_value[1]:
                        met_height_value[1] = True
                        height_min[1] -= 1

                    if met_height_value[0] and not (met_height_value[1] and height_min[1] > x_dict[str(x)][len(x_dict[str(x)]) - 1]):
                        popped = x_dict[str(x)].pop()
                        pixel_stripes[stripe_ind].append([x, popped])

        print(f'Length of 0 stripe is {len(pixel_stripes[0])}, length of {len(pixel_stripes) - 1} stripe is {len(pixel_stripes[len(pixel_stripes) - 1])}')

        return pixel_stripes

    def swap_positions(self, list, pos1, pos2):
        list[pos1], list[pos2] = list[pos2], list[pos1]
        return list



from tlcr.Video import Video
from tlcr.Line import Line
from tlcr.ImageNet import ImageNet
from skimage.util import img_as_float
import numpy as np


class Tlcr:
    intensity_k = 0.155

    @staticmethod
    def get_params(url, weights, coords):
        images = []
        Video.fill_images(url, images)
        line = Line(coords)
        images = images[::5]
        size = len(images)
        x = np.zeros((size, ImageNet.size_img, ImageNet.size_img, 3), dtype='float32')
        for i in range(0, len(images)):
            x[i] = np.array(img_as_float(line.get_image(images[i])))
        image_net = ImageNet()
        image_net.init()
        image_net.load_weights(weights)
        predictions = image_net.predict(x)
        size_predictions = len(predictions)
        tlcr = 0
        vehicles = 0
        wait_next_vehicle = True
        for i in range(0, size_predictions):
            local_tlcr = line.get_tlcr(predictions[i])
            if local_tlcr > Tlcr.intensity_k and wait_next_vehicle:
                vehicles += 1
            wait_next_vehicle = local_tlcr < Tlcr.intensity_k
            tlcr += local_tlcr
        tlcr = tlcr / size_predictions
        return tlcr, vehicles







import os
import cv2
from tlcr.Video import Video


class Bioinspired:
    def __init__(self, width, height, path_to_save):
        self.__retina = cv2.bioinspired_Retina.create((width, height))
        self.__retina.write('retinaParams.xml')
        self.__retina.setup('retinaParams.xml')
        self.__orig = 'orig'
        self.__res = 'res'
        self.__path_orig = os.path.join(path_to_save, self.__orig)
        self.__path_res = os.path.join(path_to_save, self.__res)
        if not os.path.exists(self.__path_orig):
            os.makedirs(self.__path_orig)
        if not os.path.exists(self.__path_res):
            os.makedirs(self.__path_res)

    def run(self, url):
        path_orig = os.path.join(self.__path_orig, os.path.splitext(os.path.basename(url))[0])
        path_res = os.path.join(self.__path_res, os.path.splitext(os.path.basename(url))[0])
        images = []
        Video.fill_images(url, images)
        images = images[::5]
        size = len(images)

        for i in range(0, size):
            frame = images[i]
            self.__retina.run(frame)
            res = self.__retina.getMagno()
            frame_path = '{}_{}.jpg'.format(path_orig, i)
            res_path = '{}_{}.jpg'.format(path_res, i)
            print(frame_path, res_path)
            if not cv2.imwrite(frame_path, frame):
                break
            cv2.imwrite(res_path, res)




# required to work on GPU (do not change because it only works like that)
from __future__ import absolute_import, division, print_function, unicode_literals
import tensorflow as tf
import numpy as np
import os
from tensorflow.keras import Model
from tensorflow.keras.optimizers.legacy import Adam
from tensorflow.keras.losses import binary_crossentropy
from tensorflow.keras.layers import Input, Conv2D, Conv2DTranspose, MaxPooling2D, concatenate, Dropout
from tensorflow.keras import backend as K
from tensorflow.python.keras.utils.generic_utils import get_custom_objects


class ImageNet:
    size_img = 256

    def __init__(self):
        # config = tf.compat.v1.ConfigProto(gpu_options=tf.GPUOptions(per_process_gpu_memory_fraction=0.7),
        #                                   allow_soft_placement=True)
        # config.gpu_options.allow_growth = True
        # # config.gpu.set_per_process_memory_fraction(0.4)
        # config.gpu_options.per_process_gpu_memory_fraction = 0.9
        # set_session(tf.compat.v1.Session(config=config))

        # config = tf.compat.v1.ConfigProto(gpu_options=tf.GPUOptions(per_process_gpu_memory_fraction=0.7),
        #                                   allow_soft_placement=True)
        # config.gpu_options.allow_growth = True
        # # config.gpu.set_per_process_memory_fraction(0.4)
        # set_session(tf.compat.v1.Session(config=config))
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
        self.init()

    def build_model(self, input_layer, start_neurons):
        conv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(input_layer)
        conv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(conv1)
        pool1 = MaxPooling2D((2, 2))(conv1)
        pool1 = Dropout(0.25)(pool1)

        conv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(pool1)
        conv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(conv2)
        pool2 = MaxPooling2D((2, 2))(conv2)
        pool2 = Dropout(0.5)(pool2)

        conv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(pool2)
        conv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(conv3)
        pool3 = MaxPooling2D((2, 2))(conv3)
        pool3 = Dropout(0.5)(pool3)

        conv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(pool3)
        conv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(conv4)
        pool4 = MaxPooling2D((2, 2))(conv4)
        pool4 = Dropout(0.5)(pool4)

        # Middle
        convm = Conv2D(start_neurons * 16, (3, 3), activation="relu", padding="same")(pool4)
        convm = Conv2D(start_neurons * 16, (3, 3), activation="relu", padding="same")(convm)

        deconv4 = Conv2DTranspose(start_neurons * 8, (3, 3), strides=(2, 2), padding="same")(convm)
        uconv4 = concatenate([deconv4, conv4])
        uconv4 = Dropout(0.5)(uconv4)
        uconv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(uconv4)
        uconv4 = Conv2D(start_neurons * 8, (3, 3), activation="relu", padding="same")(uconv4)

        deconv3 = Conv2DTranspose(start_neurons * 4, (3, 3), strides=(2, 2), padding="same")(uconv4)
        uconv3 = concatenate([deconv3, conv3])
        uconv3 = Dropout(0.5)(uconv3)
        uconv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(uconv3)
        uconv3 = Conv2D(start_neurons * 4, (3, 3), activation="relu", padding="same")(uconv3)

        deconv2 = Conv2DTranspose(start_neurons * 2, (3, 3), strides=(2, 2), padding="same")(uconv3)
        uconv2 = concatenate([deconv2, conv2])
        uconv2 = Dropout(0.5)(uconv2)
        uconv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(uconv2)
        uconv2 = Conv2D(start_neurons * 2, (3, 3), activation="relu", padding="same")(uconv2)

        deconv1 = Conv2DTranspose(start_neurons * 1, (3, 3), strides=(2, 2), padding="same")(uconv2)
        uconv1 = concatenate([deconv1, conv1])
        uconv1 = Dropout(0.5)(uconv1)
        uconv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(uconv1)
        uconv1 = Conv2D(start_neurons * 1, (3, 3), activation="relu", padding="same")(uconv1)

        uncov1 = Dropout(0.5)(uconv1)
        output_layer = Conv2D(1, (1, 1), padding="same", activation="sigmoid")(uconv1)
        return output_layer

    def dice_coeficient(self, y_true, y_pred):
        y_true_f = K.flatten(y_true)
        y_pred = K.cast(y_pred, 'float32')
        y_pred_f = K.cast(K.greater(K.flatten(y_pred), 0.5), 'float32')
        intersection = y_true_f * y_pred_f
        score = 2. * K.sum(intersection) / (K.sum(y_true_f) + K.sum(y_pred_f))
        return score

    def dice_loss(self, y_true, y_pred):
        smooth = 1.
        y_true_f = K.flatten(y_true)
        y_pred_f = K.flatten(y_pred)
        intersection = y_true_f * y_pred_f
        score = (2. * K.sum(intersection) + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)
        return 1. - score

    def bce_dice_loss(self, y_true, y_pred):
        return binary_crossentropy(y_true, y_pred) + self.dice_loss(y_true, y_pred)

    def get_iou_vector(self, A, B):
        # Numpy version

        batch_size = A.shape[0]
        metric = 0.0
        for batch in range(batch_size):
            t, p = A[batch], B[batch]
            true = np.sum(t)
            pred = np.sum(p)

            # deal with empty mask first
            if true == 0:
                metric += (pred == 0)
                continue

            # non empty mask case.  Union is never empty
            # hence it is safe to divide by its number of pixels
            intersection = np.sum(t * p)
            union = true + pred - intersection
            iou = intersection / union

            # iou metric is a stepwise approximation of the real iou over 0.5
            iou = np.floor(max(0, (iou - 0.45) * 20)) / 10

            metric += iou

        # take the average over all images in batch
        metric /= batch_size
        return metric

    def my_iou_metric(self, label, pred):
        # Tensorflow version
        return tf.py_func(self.get_iou_vector, [label, pred > 0.5], tf.float64)

    def init(self):
        get_custom_objects().update({'bce_dice_loss': self.bce_dice_loss})
        get_custom_objects().update({'dice_loss': self.dice_loss})
        get_custom_objects().update({'dice_coef': self.dice_coeficient})
        get_custom_objects().update({'my_iou_metric': self.my_iou_metric})

        input_layer = Input((self.size_img, self.size_img, 3))
        output_layer = self.build_model(input_layer, 16)
        self.model = Model(input_layer, output_layer)
        self.model.compile(loss=self.bce_dice_loss, optimizer=Adam(lr=1e-3), metrics=[self.my_iou_metric])

    def train(self, train_x, train_y, epochs=10):
        history = self.model.fit(train_x, train_y, batch_size=32, epochs=epochs, verbose=1, validation_split=0.1)

    def save_weights(self, file_name):
        self.model.save_weights(file_name)

    def load_weights(self, file_name):
        self.model.load_weights(file_name, by_name=False)

    def predict(self, image_array):
        return self.model.predict(image_array)



from setuptools import setup

setup(
    name='tlcr',
    version='0.0.1',
    description='tlcr',
    license='MIT',
    packages=['tlcr'],
    author='Mykola Onyshchuk',
    author_email='mykolaon1@gmail.com',
    keywords=['tlcr', 'intensity'],
    url='https://github.com/MykolaOnyshchuk/tlcr_intensity_lib'
)
